from torch.utils import data
import gzip
import sys
import torch.optim as optim
import torch.nn as nn
import torch.optim.lr_scheduler
import torch.nn.init

from torch.nn import Module, Sequential, Conv3d
import globalvar as gl
import config



class_number = gl.get_value('class_number')
num_layers = gl.get_value('num_layers')
kernel_size = gl.get_value('kernel_size')
batch_size = gl.get_value('batch_size')
learning_rate = gl.get_value('learning_rate')
growth_rate = gl.get_value('growth_rate')
dilation_mod =  gl.get_value('dilation_mod')
log_dir = gl.get_value('log_dir')
modelsave_dir = gl.get_value('modelsave_dir')
epochs = gl.get_value('epochs')
device_ids = gl.get_value('device_ids')


def count_parameters(module: Module):
    return sum(p.numel() for p in module.parameters() if p.requires_grad)


def count_conv3d(module: Module):
    return len([m for m in module.modules() if isinstance(m, Conv3d)])


'''
3D MS-D

MixedScaleDenseLayer is the basic layer, functions like Conv2d, 
however, its input feature map number is linearly increasing as
net depth increases;

Net class is used to build a MS-D net
num_layers: depth of MS-D net
growth_rate: width in MS-D net paper, that is w
kernel_size: in MS-D, there is only one kind of kernel, such as 3*3*3
dilation_mod: in MS-D, every ten layers combine a weak block, we will talk it
              in detail in readme.txt

'''



class MixedScaleDenseLayer(Module):
    def __init__(self, in_channels, dilations):
        super(MixedScaleDenseLayer, self).__init__()

        if type(dilations) == int:
            dilations = [j % dilation_mod + 1 for j in range(dilations)]

        self.kernel_size = kernel_size
        self.in_channels = in_channels
        self.out_channels = in_channels + len(dilations)

        for j, dilation in enumerate(dilations):
            # Equal to: kernel_size + (kernel_size - 1) * (dilation - 1)
            dilated_kernel_size = (kernel_size - 1) * dilation + 1
            padding = dilated_kernel_size // 2
            self.add_module(f'conv_{j}', Conv3d(
                in_channels, 1,
                kernel_size=kernel_size, dilation=dilation, padding=padding
            ))


    def forward(self, x):
        return torch.cat((x,) + tuple(c(x) for c in self.children()), dim=1)


class Net(Sequential):
    def __init__(self, in_channels=1, out_channels=class_number):
        super(Net, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        #size_diff = 13
        current_channels = in_channels
        for i in range(num_layers):
            dilations = [((i * growth_rate + j) % dilation_mod) +
                         1 for j in range(growth_rate)]
            l = MixedScaleDenseLayer(current_channels, dilations)
            current_channels = l.out_channels
            self.add_module(f'layer_{i}', l)      
            self.add_module(f'relu_{i}',torch.nn.LeakyReLU())  

        self.add_module('last', Conv3d(current_channels, out_channels, kernel_size=1))
        #self.add_module('last_activation', torch.nn.LeakyReLU())
        self.add_module('last_softmax', torch.nn.LogSoftmax(dim=1))
      

       


